# +
import rasterio
import numpy as np

def sma(image, endmembers, nodata = -99999):
    
    '''
    The SMA assumes that the energy received within the field of vision of the remote sensor 
    can be considered as the sum of the energies received from each dominant endmember. 
    This function addresses a Linear Mixing Model.
    
    A regression analysis is used to obtain the fractions. In least squares inversion algorithms, 
    the common objective is to estimate abundances that minimize the squared error between the 
    actual spectrum and the estimated spectrum. The values of the fractions will be between 0 and 1.
    
    Parameters:
    
        image: Optical images. It must be rasterio.io.DatasetReader with 3d.
        
        endmembers: Endmembers must be a matrix (numpy.ndarray) and with more than one endmember. 
                    Rows represent the endmembers and columns represent the spectral bands.
                    The number of bands must be greater than the number of endmembers.
                    E.g. an image with 6 bands, endmembers dimension should be $n*6$, where $n$ 
                    is rows with the number of endmembers and 6 is the number of bands 
                    (should be equal).
                    
        nodata: The NoData value to replace with -99999.
    
    Return:
    
        numpy.ndarray with 2d.
        
    References:
    
        Adams, J. B., Smith, M. O., & Gillespie, A. R. (1993). Imaging spectroscopy:
        Interpretation based on spectral mixture analysis. In C. M. Pieters & P.
        Englert (Eds.), Remote geochemical analysis: Elements and mineralogical
        composition. NY: Cambridge Univ. Press 145-166 pp.
    
        Shimabukuro, Y.E. and Smith, J., (1991). The least squares mixing models to
        generate fraction images derived from remote sensing multispectral data.
        IEEE Transactions on Geoscience and Remote Sensing, 29, pp. 16-21.
    
    Note:
    A regression analysis is used to obtain the fractions. In least squares
    inversion algorithms, the common objective is to estimate abundances that
    minimize the squared error between the actual spectrum and the estimated spectrum.
    The values of the fractions will be between 0 and 1.
    '''
    
    if not isinstance(image, (rasterio.io.DatasetReader)):
        raise TypeError('"image" must be raster read by rasterio.open().')
        
    if not isinstance (endmembers, (np.ndarray)):
        raise ErrorType('"endmembers" must be numpy.ndarray.')
    
    bands = image.count
        
    rows = image.height
        
    cols = image.width
    
    # number of endmembers
    n_endm = endmembers.shape[0]
    
    # number of bands extracted for each endmember
    b_endm = endmembers.shape[1]
        
    st = image.read()
        
    # data in [rows, cols, bands]
    st_reorder = np.moveaxis(st, 0, -1) 
    # data in [rows*cols, bands]
    arr = st_reorder.reshape((rows*cols, bands))
    
    # nodata
    if np.isnan(np.sum(arr)):
        arr[np.isnan(arr)] = self.nodata
    
    if not arr.shape[1] > n_endm:
        raise ValueError('The number of bands must be greater than the number of endmembers.')
    
    if not arr.shape[1] == b_endm:
        raise ValueError('The number of values extracted in band should be equal.')
    
    M = np.transpose(endmembers)
    
    mat_oper = np.dot(np.linalg.inv(np.dot(np.transpose(M), M)), np.transpose(M)) 
    
    frac = np.zeros((rows*cols, n_endm))
                
    for i in np.arange(0, n_endm, 1):
        for j in np.arange(0, rows*cols, 1):
            f = np.dot(mat_oper, arr[j,:])
            frac[j,i] = f[i,]
    
    sma_img = frac.reshape((rows, cols, n_endm))
    
    return sma_img



# -*- coding: utf-8 -*-
# +
from sklearn.cluster import KMeans
import rasterio
import numpy as np

def calkmeans(image, k = None, algo = ("auto", "elkan"), max_iter = 300, n_iter = 10, nodata = -99999, **kwargs):
    
    '''
    Calibrating kmeans
    
    This function allows to calibrate the kmeans algorithm. It is possible to obtain
    the best 'k' value and the best embedded algorithm in KMmeans.
        
    Parameters:
        image: Optical images. It must be rasterio.io.DatasetReader with 3d.
            
        k: k This argument is None when the objective is to obtain the best 'k' value. 
           f the objective is to select the best algorithm embedded in kmeans, please specify a 'k' value.
    
        max_iter: The maximum number of iterations allowed. Strictly related to KMeans. Please see
                  https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html
            
        algo: It can be "auto" and 'elkan'. "auto" and "full" are deprecated and they will be 
              removed in Scikit-Learn 1.3. They are both aliases for "lloyd".
            
        Changed in version 1.1: Renamed “full” to “lloyd”, and deprecated “auto” and “full”. 
                               Changed “auto” to use “lloyd” instead of “elkan”.
            
        n_iter: Iterations number to obtain the best 'k' value. 'n_iter' must be greater than the 
                number of classes expected to be obtained in the classification. Default is 10.
            
        nodata: The NoData value to replace with -99999. 
                     
        **kwargs: These will be passed to scikit-learn KMeans, please see full lists at:
                  https://scikit-learn.org/stable/modules/generated/sklearn.cluster.KMeans.html.
    
        Return:
            Labels of classification as numpy object with 2d.
            
        Note:
        If the idea is to find the optimal value of 'k' (clusters or classes), k = None as an 
        argument of the function must be put, because the function find 'k' for which the intra-class 
        inertia is stabilized. If the 'k' value is known and the idea is to find the best algorithm 
        embedded in kmeans (that maximizes inter-class distances), k = n, which 'n' is a specific 
        class number, must be put. It can be greater than or equal to 0.  
        
    '''
    
    if not isinstance(image, (rasterio.io.DatasetReader)):
        raise TypeError('"image" must be raster read by rasterio.open().')
        
    bands = image.count
        
    rows = image.height
        
    cols = image.width
        
    # stack images
    #l = []
    #for i in np.arange(1, bands+1): l.append(image.read(int(i)))
    #st = np.stack(l)
    st = image.read()
    
    # data in [rows, cols, bands]
    st_reorder = np.moveaxis(st, 0, -1) 
    # data in [rows*cols, bands]
    arr = st_reorder.reshape((rows*cols, bands))
    
    # nodata
    if np.isnan(np.sum(arr)):
        arr[np.isnan(arr)] = self.nodata
    
    if k is None:
        
        if 'auto' in algo:
            
            inertias_intra_lloyd_k = []
    
            for i in range(1, n_iter + 1):
    
                # Building and fitting the model
                kmeanModel = KMeans(n_clusters = i, max_iter = max_iter, algorithm = 'auto', **kwargs)
    
                kmeanModel.fit(arr)

                inertias_intra_lloyd_k.append(kmeanModel.inertia_)
        
        if 'elkan' in algo:
            
            inertias_intra_elkan_k = []
    
            for i in range(1, n_iter + 1):
    
                # Building and fitting the model
                kmeanModel = KMeans(n_clusters = i, max_iter = max_iter, algorithm = 'elkan', **kwargs)
    
                kmeanModel.fit(arr)

                inertias_intra_elkan_k.append(kmeanModel.inertia_)
        
        dic = {'inertias_intra_lloyd_k': inertias_intra_lloyd_k,
               'inertias_intra_elkan_k': inertias_intra_elkan_k}

        all_algorithms = ("auto", "elkan")

        for model in all_algorithms:
            if model not in algo:
                del dic[model]
                    
        find_k = dic
        
        return find_k
    
    # nc = range(1,10)
    # kmeanlist = [KMeans(n_clusters = i) for i in Nc]
    # variance = [kmeanlist[i].fit(datos).inertia_ for i in range(len(kmeanlist))]
    
    elif isinstance(k, (int)):
        
        if 'auto' in algo:
            
            inertias_inter_lloyd = []
    
            for i in range(n_iter):
    
                kmeanModel = KMeans(n_clusters = k, max_iter = max_iter, algorithm = 'auto', **kwargs)
    
                kmeanModel.fit(arr)

                inertias_inter_lloyd.append(kmeanModel.score(arr))
        
        if 'elkan' in algo:
            
            inertias_inter_elkan = []
    
            for i in range(n_iter):
    
                kmeanModel = KMeans(n_clusters = k, max_iter = max_iter, algorithm = 'elkan', **kwargs)
    
                kmeanModel.fit(arr)

                inertias_inter_elkan.append(kmeanModel.inertia_)
        
        dic = {'inertias_inter_lloyd': inertias_inter_lloyd,
               'inertias_inter_elkan': inertias_inter_elkan}

        all_algorithms = ("auto", "elkan")

        for model in all_algorithms:
            if model not in algo:
                del dic[model]
                    
        best_algorithm = dic
        
        return best_algorithm
    
    else:
        raise TypeError(f'{k} must be None or positive integer. type = {type(k)}')



# -*- coding: utf-8 -*-
# +
import os
import glob
import numpy as np
import rasterio

class atmosCorr(object):
    
    '''Atmospheric Correction in Optical domain'''
    
    def __init__(self, path, nodata = -99999):
        
        '''
        
        Parameter:
        
            path: String. The folder in which the satellite bands are located. This images could be Landsat
                  Collection 2 Level-1. For example: path = r'/folder/image/raster'.
            
            nodata: The NoData value to replace with -99999.
        '''
        
        self.path = path
        self.nodata = nodata
        
        path = path.replace(os.sep, '/')
        
        path = os.chdir(path)
        
        file_mtl = glob.glob("*_MTL.txt")

        dict_mtl = {}

        with open(file_mtl[0], 'r') as mtl:
            for lines in mtl:
                lines = lines.strip()
                if lines != 'END':
                    key, value = lines.split('=')
                    dict_mtl[key] = value

        names_bands = glob.glob("*B?*.TIF")
        names_bands.sort()
        
        self.dict_mtl = dict_mtl
        self.names_bands = names_bands

    def RAD(self, sat = 'LC08'):
        
        '''
        Conversion to TOA Radiance. Landsat Level-1 data can be converted to TOA spectral radiance 
        using the radiance rescaling factors in the MTL file:
        
        Lλ = MLQcal + AL 
        
        where:

        Lλ = TOA spectral radiance (Watts/(m2*srad*μm))
        ML = Band-specific multiplicative rescaling factor from the metadata (RADIANCE_MULT_BAND_x, where x is the band number)
        AL = Band-specific additive rescaling factor from the metadata (RADIANCE_ADD_BAND_x, where x is the band number)
        Qcal =  Quantized and calibrated standard product pixel values (DN) 

        Parameters:
            sat: Type of Satellite. It could be Landsat-5 TM, Landsat-8 OLI or Landsat-9 OLI-2.
        
        Return: 
            An array with radiance values with 3d, i.e. (rows, cols, bands).
        '''
        
        if sat == 'LC09':
            # Landsat-9 bands -> blue, green, red, nir, swir1 and swir2
            bands_lc09 = ['B2.','B3.','B4.','B5.','B6.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lc09)]
        
        if sat == 'LC08':
            # Landsat-8 bands -> blue, green, red, nir, swir1 and swir2
            bands_lc08 = ['B2.','B3.','B4.','B5.','B6.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lc08)]
        
        if sat == 'LT05':
            # Landsat-5 bands -> blue, green, red, nir, swir1 and swir2
            bands_lt05 = ['B1.','B2.','B3.','B4.','B5.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lt05)]

        list_bands = []
        for i in output_names:
            with rasterio.open(i, 'r') as b:
                list_bands.append(b.read(1))

        dn = np.stack(list_bands)

        # data in [rows, cols, bands]
        #dn = np.moveaxis(st, 0, -1) 
        
        rows = dn.shape[1]
        
        cols = dn.shape[2]

        bands = dn.shape[0]

        # nodata
        if np.isnan(np.sum(dn)):
            dn[np.isnan(dn)] = self.nodata
        
        tetha = float(self.dict_mtl['SUN_ELEVATION '])
        
        rad_bands = []
        
        for i in range(bands):
            
            ML = float(self.dict_mtl['RADIANCE_MULT_BAND_' + str(i+1) + ' '])
            
            AL = float(self.dict_mtl['RADIANCE_ADD_BAND_' + str(i+1) + ' '])
            
            radiance = np.add(np.multiply(dn[i,:,:], ML), AL) # Lλ = ML*DN + AL 
            
            rad_bands.append(radiance)
        
        arr_rad = np.moveaxis(np.stack(rad_bands), 0, -1) 
        
        # negative values -> nodata
        arr_rad[arr_rad < 0] = np.nan
        
        return arr_rad
    
    def TOA(self, sat = 'LC08'):
        
        '''
        A reduction in scene-to-scene variability can be achieved by converting the at-sensor 
        spectral radiance to exoatmospheric TOA reflectance, also known as in-band planetary albedo.
        
        Equation to obtain TOA reflectance:
        
        ρλ′ = Mρ*DN + Aρ
        
        ρλ = ρλ′/sin(theta)
        
        Parameters:
            sat: Type of Satellite. It could be Landsat-5 TM, Landsat-8 OLI or Landsat-9 OLI-2.
        
        Return: 
            An array with TOA values with 3d, i.e. (rows, cols, bands).
        '''
        
        if sat == 'LC09':
            # Landsat-9 bands -> blue, green, red, nir, swir1 and swir2
            bands_lc09 = ['B2.','B3.','B4.','B5.','B6.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lc09)]
        
        if sat == 'LC08':
            # Landsat-8 bands -> blue, green, red, nir, swir1 and swir2
            bands_lc08 = ['B2.','B3.','B4.','B5.','B6.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lc08)]
        
        if sat == 'LT05':
            # Landsat-5 bands -> blue, green, red, nir, swir1 and swir2
            bands_lt05 = ['B1.','B2.','B3.','B4.','B5.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lt05)]

        list_bands = []
        for i in output_names:
            with rasterio.open(i, 'r') as b:
                list_bands.append(b.read(1))

        dn = np.stack(list_bands)

        # data in [rows, cols, bands]
        #dn = np.moveaxis(st, 0, -1) 
        
        rows = dn.shape[1]
        
        cols = dn.shape[2]

        bands = dn.shape[0]

        # nodata
        if np.isnan(np.sum(dn)):
            dn[np.isnan(dn)] = self.nodata
        
        tetha = float(self.dict_mtl['SUN_ELEVATION '])
        
        toa_bands = []
        
        for i in range(bands):
            
            Mp = float(self.dict_mtl['REFLECTANCE_MULT_BAND_' + str(i+1) + ' '])
            
            Ap = float(self.dict_mtl['REFLECTANCE_ADD_BAND_' + str(i+1) + ' '])
            
            plambda = np.add(np.multiply(dn[i,:,:], Mp), Ap) # ρλ′ = Mρ*DN + Aρ
            
            TOA = plambda/np.sin((tetha*np.pi/180)) # ρλ = ρλ′/sin(theta)
            
            toa_bands.append(TOA)
        
        arr_toa = np.moveaxis(np.stack(toa_bands), 0, -1) 
        
        # negative values -> nodata
        arr_toa[arr_toa < 0] = np.nan
        
        return arr_toa
    
    def DOS(self, sat = 'LC08', mindn = None):
        
        '''
        The Dark Object Subtraction Method was proposed by Chavez (1988). This image-based 
        atmospheric correction method considers absolutely critical and valid the existence 
        of a dark object in the scene, which is used in the selection of a minimum value in 
        the haze correction. The most valid dark objects in this kind of correction are areas 
        totally shaded or otherwise areas representing dark water bodies.
        
        Parameters:
        
            sat: Type of Satellite. It could be Landsat-5 TM, Landsat-8 OLI or Landsat-9 OLI-2.
            
            mindn: Min of digital number for each band in a list.
            
        Return: 
            An array with Surface Reflectance values with 3d, i.e. (rows, cols, bands).
        
        References:
        
            Chavez, P.S. (1988). An Improved Dark-Object Subtraction Technique for Atmospheric 
            Scattering Correction of Multispectral Data. Remote Sensing of Envrironment, 24(3), 459-479.
        '''
        
        if sat == 'LC09':
            bands_lc09 = ['B2.','B3.','B4.','B5.','B6.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lc09)]
            # ESUN for Landsat-9 bands -> blue, green, red, nir, swir1 and swir2
            # https://www.usgs.gov/landsat-missions/using-usgs-landsat-level-1-data-product
            ESUN = [2067, 1893, 1603, 972.6, 245.0, 79.72]
            
        if sat == 'LC08':
            bands_lc08 = ['B2.','B3.','B4.','B5.','B6.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lc08)]
            # ESUN for Landsat-8 bands -> blue, green, red, nir, swir1 and swir2
            # https://www.usgs.gov/landsat-missions/using-usgs-landsat-level-1-data-product
            ESUN = [2067, 1893, 1603, 972.6, 245.0, 79.72]

        if sat == 'LT05':
            # Landsat-5 bands -> blue, green, red, nir, swir1 and swir2
            bands_lt05 = ['B1.','B2.','B3.','B4.','B5.','B7.']
            output_names = [name for name in self.names_bands if (name[41:44] in bands_lt05)]
            
        list_bands = []
        for i in output_names:
            with rasterio.open(i, 'r') as b:
                list_bands.append(b.read(1))

        dn = np.stack(list_bands)

        # data in [rows, cols, bands]
        #dn = np.moveaxis(st, 0, -1) 
        
        rows = dn.shape[1]
        
        cols = dn.shape[2]

        bands = dn.shape[0]

        # Min of digital number for each band
        if mindn is None:
            dn_mins = [np.nanmin(dn[i,:,:][np.nonzero(dn[i,:,:])]) for i in range(bands)]
        else:
            dn_mins = mindn
            
            if not len(mindn) == bands:
                raise ValueError(f'The length of the "mindn" argument must be equal to the number'
                                f'of bands. Length of "mindn" is {len(mindn)}, and length of bands'
                                f'is {bands}')
                
        # nodata
        if np.isnan(np.sum(dn)):
            dn[np.isnan(dn)] = self.nodata
        
        tetha = float(self.dict_mtl['SUN_ELEVATION '])
        u_a = float(self.dict_mtl['EARTH_SUN_DISTANCE '])
        
        dos_bands = []
        
        for i in range(bands):
            
            ML = float(self.dict_mtl['RADIANCE_MULT_BAND_' + str(i+1) + ' '])
            
            AL = float(self.dict_mtl['RADIANCE_ADD_BAND_' + str(i+1) + ' '])
            
            L = np.add(np.multiply(dn[i,:,:], ML), AL) # L = ML*DN+AL
            
            L_min = np.add(np.multiply(dn_mins[i], ML), AL) # for each band
            
            DOS = np.pi*(np.multiply((L - L_min), (u_a)**2))/(ESUN[i]*np.sin((tetha*np.pi/180)))
            
            dos_bands.append(DOS)
        
        arr_dos = np.moveaxis(np.stack(dos_bands), 0, -1) 
        
        # negative values -> nodata
        arr_dos[arr_dos < 0] = np.nan
        
        return arr_dos



# +
import rasterio
import numpy as np
import os

def writeRaster(arr, image, filename = None, filepath = None, n = 1):
    
    '''
    This algorithm allows to save array images to raster format (.tif).
    
    Parameters:
    
        arr: Array object with 2d (rows and cols) or 3d (rows, cols, bands).
        
        image: Optical images. It must be read by rasterio.open().
        
        filename: The image name to be saved.
        
        filepath: The path which the image will be stored.
        
        n: Number of images to be saved.
    
    Return:
        A raster in your filepath.
    '''
    
    if not isinstance(arr, (np.ndarray)):
        raise TypeError(f'{arr} must be an array with 2d (rows and cols).')
    
    if not isinstance(image, (rasterio.io.DatasetReader)): 
        raise TypeError(f'The image must be rasterio.io.DatasetReader. image = {type(image)}')
    
    if not arr.ndim >= 2:
        raise ValueError(f'Input mus have two or three dimensions. ndim = {arr.ndim}')
    
    if filename is None:
        
        filename = 'raster'
    
    if filepath is None:
        
        current_path = os.getcwd()
        
        path = current_path.replace(os.sep, '/') + '/' + filename + '.tif'
        
    else:
        path = filepath.replace(os.sep, '/') + '/' + filename + '.tif'
    
    path_name = path
    
    if arr.ndim == 2:
        n = 1
    else:
        n = arr.shape[2] # number of bands >= 2
        
    kwargs = image.meta
    
    if arr.dtype == 'int64':
        arr = arr.astype(np.int32)
    
    dtype = str(arr.dtype)
    
    kwargs.update({      
        'dtype': dtype,
        'count': n
    })
    
    if arr.ndim == 2:
        with rasterio.open(path_name, 'w', compress='lzw', **kwargs) as img:    
            img.write_band(1, arr)
    else:
        with rasterio.open(path_name, 'w', compress='lzw', **kwargs) as img:
            for num, name in enumerate([arr[:,:,i] for i in range(n)], start=1):
                img.write_band(num, name)



# +
import copy
import rasterio
import numpy as np
import pandas as pd
from dbfread import DBF
from tensorflow import keras
from keras import models
from keras import layers
from keras.utils import to_categorical
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics import confusion_matrix

class DL(object):
    
    'Deep Learning classification in Remote Sensing'
    
    def __init__(self, image, endmembers, nodata = -99999):
        
        '''
        Parameter:
        
            image: Optical images. It must be rasterio.io.DatasetReader with 3d.
            
            endmembers: Endmembers must be a matrix (numpy.ndarray) and with more than one endmember. 
                    Rows represent the endmembers and columns represent the spectral bands.
                    The number of bands must be equal to the number of endmembers.
                    E.g. an image with 6 bands, endmembers dimension should be $n*6$, where $n$ 
                    is rows with the number of endmembers and 6 is the number of bands 
                    (should be equal).
                    In addition, Endmembers must have a field (type int or float) with the names 
                    of classes to be predicted.
            
            nodata: The NoData value to replace with -99999.
            
        '''
        
        self.image = image
        self.endm = endmembers
        self.nodata = nodata
        
        if not isinstance(self.image, (rasterio.io.DatasetReader)):
            raise TypeError('"image" must be raster read by rasterio.open().')
        
        bands = self.image.count
        
        rows = self.image.height
        
        cols = self.image.width
        
        st = image.read()
        
        # data in [rows, cols, bands]
        st_reorder = np.moveaxis(st, 0, -1) 
        # data in [rows*cols, bands]
        arr = st_reorder.reshape((rows*cols, bands))

        # nodata
        #if np.isnan(np.sum(arr)):
            #arr[np.isnan(arr)] = self.nodata
        
        # dealing with nan
        key_nan = np.isnan(np.sum(arr[:,0]))
        
        if key_nan:
            # saving an array for predicted classes
            class_final = arr[:, 0].copy()
            # positions with nan
            posIndx = np.argwhere(~np.isnan(class_final)).flatten()
            # replace np.nan -> 0
            arr[np.isnan(arr)] = 0
            
        # if it is read by pandas.read_csv()
        if isinstance(self.endm, (pd.core.frame.DataFrame)):
            
            for i in np.arange(self.endm.shape[1]):
                
                if all(self.endm.iloc[:,int(i)] < 100) & all(self.endm.iloc[:,int(i)] >= 0): indx = i; break
        
        # if the file is .dbf    
        elif isinstance(self.endm, (DBF)): # isinstance() function With Inheritance
            
            self.endm = pd.DataFrame(iter(self.endm))
            
            for i in np.arange(self.endm.shape[1]):
                
                if all(self.endm.iloc[:,int(i)] < 100) & all(self.endm.iloc[:,int(i)] >= 0): indx = i; break
        
        else:
            raise TypeError('"endm" must be .csv (pandas.core.frame.DataFrame).')

        if not self.endm.shape[1] == (bands + 1):
            raise ValueError('The number of columns of signatures (included the class column) must'
                             'be equal to the number of bands + 1.')
        
        self.indx = indx
        
        self.arr = arr
        
        self.rows = rows
        
        self.cols = cols
        
        if key_nan:
            self.key_nan = key_nan
            self.posIndx = posIndx
            self.class_final = class_final
        else:
            self.key_nan = key_nan
        
    def FullyConnected(self, hidden_layers = 3, hidden_units = [64, 32, 16], output_units = 10,
                       input_shape = (6,), epochs = 300, batch_size = 32, training_split = 0.8, 
                       random_state = None):
        
        ''' 
        This algorithm consists of a network with a sequence of Dense layers, which are densely 
        connected (also called *fully connected*) neural layers. This is the simplest of deep 
        learning.
        
        Parameters:
    
            hidden_layers: Number of hidden layers to be used. 3 is the default.
    
            hidden_units: Number of units to be used. This is related to 'neurons' in each hidden 
                          layers.
                          
            output_units: Number of clases to be obtained.
            
            input_shape: The input shape is generally the shape of the input data provided to the 
                         Keras model while training. The model cannot know the shape of the 
                         training data. The shape of other tensors(layers) is computed automatically.
            
            epochs: Number of iterations, the network will compute the gradients of the weights with
                    regard to the loss on the batch, and update the weights accordingly.
            
            batch_size: This break the data into small batches. In deep learning, models do not 
                        process antire dataset at once.
            
            training_split: For splitting samples into two subsets, i.e. training data and for testing
                            data.
            
            random_state: Random state ensures that the splits that you generate are reproducible. 
                          Please, see for more details https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.train_test_split.html
            
        Return:
        
            A dictionary with Labels of classification as numpy object, overall accuracy, 
            among others results.
        '''
        
        # removing the class column
        X = self.endm.drop(self.endm.columns[[self.indx]], axis = 1)
        
        # only predictor variables
        y = self.endm.iloc[:, self.indx]
        
        # split in training and testing
        Xtrain, Xtest, ytrain, ytest = train_test_split(
            X, 
            y, 
            train_size = training_split, 
            test_size = 1 - training_split, 
            random_state = random_state)
        
        ytrain_catego = to_categorical(ytrain)
        ytest_catego = to_categorical(ytest)
        
        fcl = [layers.Dense(hidden_units[0], activation = 'relu', input_shape = input_shape)]
        
        for i in np.arange(1, hidden_layers):
            
            fcl.append(layers.Dense(hidden_units[i], activation = 'relu'))
        
        fcl.append(layers.Dense(output_units, activation = 'softmax'))
        
        model = models.Sequential(fcl)
        
        model.compile(optimizer = 'Adam', loss = 'categorical_crossentropy', metrics = ['accuracy'])
        
        # model trained
        model.fit(Xtrain, ytrain_catego, epochs = epochs, batch_size = batch_size, verbose = 0)
        
        labels_fullyconnected = model.predict(self.arr)
        
        labels = [np.argmax(element) for element in labels_fullyconnected]
        
        labels = np.array(labels)
        
        # dealing with nan
        if self.key_nan:
            # image border like nan
            labels = labels[self.posIndx]
            self.class_final[self.posIndx] = labels
            class_fullyconnected = self.class_final.reshape((self.rows, self.cols))
        else:
            class_fullyconnected = labels.reshape((self.rows, self.cols))
  
        # Confusion matrix
        predict_prob = model.predict(Xtest)
        
        predict_Xtest = [np.argmax(element) for element in predict_prob]
        
        predict_Xtest = np.array(predict_Xtest)
        
        MC = confusion_matrix(ytest, predict_Xtest)
    
        MC = np.transpose(MC)

        # Users Accuracy and Commission
        total_rows = np.sum(MC, axis = 1)
        ua = np.diag(MC)/np.sum(MC, axis = 1)*100
        co = 100 - ua

        # Producer Accuracy and Comission
        total_cols = np.sum(MC, axis = 0)
        pa = np.diag(MC)/np.sum(MC, axis = 0)*100
        om = 100 - pa

        total_cols = np.concatenate([total_cols, np.repeat(np.nan, 3)])
        pa = np.concatenate([pa, np.repeat(np.nan, 3)])
        om = np.concatenate([om, np.repeat(np.nan, 3)])

        n = MC.shape[0]
        cm = np.concatenate([MC, total_rows.reshape((n,1)), ua.reshape((n,1)), 
                             co.reshape((n,1))], axis = 1)
        cm = np.concatenate([cm, total_cols.reshape((1,n+3)), pa.reshape((1,n+3)), 
                             om.reshape((1,n+3))])

        namesCol = []
        for i in np.arange(0, n):
            stri = str(i)
            namesCol.append(stri)

        namesCol.extend(['Total', 'Users_Accuracy', 'Commission'])

        namesRow = []
        for i in np.arange(0, n):
            stri = str(i)
            namesRow.append(stri)

        namesRow.extend(['Total', 'Producer_Accuracy', 'Omission'])
        namesRow

        confusionMatrix = pd.DataFrame(cm, 
                                        columns = namesCol, 
                                        index = namesRow)

        oa = accuracy_score(ytest, predict_Xtest)
        kappa = cohen_kappa_score(ytest, predict_Xtest)
        
        output = {'Overall_Accuracy': oa,
                  'Kappa_Index': kappa,
                  'Confusion_Matrix': confusionMatrix,
                  'Classification_Map': class_fullyconnected,
                  'Image': self.image
                 } 
    
        return output
    
# -





#!/usr/bin/env python
"""The setup script."""

import io
from os import path as op
from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

here = op.abspath(op.dirname(__file__))

# get the dependencies and installs
with io.open(op.join(here, "requirements.txt"), encoding="utf-8") as f:
    all_reqs = f.read().split("\n")

install_requires = [x.strip() for x in all_reqs if "git+" not in x]
dependency_links = [x.strip().replace("git+", "") for x in all_reqs if "git+" not in x]

requirements = [ ]

setup_requirements = [ ]

test_requirements = [ ]

setup(
    author="Yonatan Tarazona",
    author_email='ytarazona09@gmail.com',
    python_requires='>=3.5',
    classifiers=[
        #'Development Status ::  2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.10'
    ],
    description="Remote Sensing Tools",
    install_requires=install_requires,
    dependency_links=dependency_links,
    license="Apache Software License 2.0",
    long_description=readme,
    long_description_content_type='text/markdown',
    include_package_data=True,
    keywords='scikeo',
    name='scikeo',
    packages=find_packages(include=['scikeo', 'scikeo.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/yotarazona/scikit-eo',
    version='0.2.34',
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# +
import rasterio
import numpy as np
import matplotlib.pyplot as plt

def plotHist(image, bands = 1, bins = 128, alpha = 0.8, title = None, xlabel = None, ylabel = None,
             label = None, ax = None, density = True, **kwargs):
    
    '''
    This function allows to plot satellite images histogram.
    
    Parameters:
        image: Optical images. It must be rasterio.io.DatasetReader with 3d or 2d.
            
        bands: Must be specified as a number of a list.
            
        bins: By default is 128. 
            
        alpha: Percentage (%) of transparency between 0 and 1. 0 indicates 0% and 1 indicates
               100%. By default is 80%.
            
        title: Assigned title.
        
        xlabel: X axis title.
        
        ylabel: Y axis title.
            
        label: Labeling the histogram.
        
        ax: current axes
        
        **kwargs: These will be passed to the matplotlib imshow(), please see full lists at:
                https://matplotlib.org/stable/api/_as_gen/matplotlib.axes.Axes.hist.html
    
    Return:
        ax: A histogram of an image.
        
    '''
    
    if not isinstance(image, (rasterio.io.DatasetReader)):
        raise TypeError('"image" must be raster read by rasterio.open().')
    
    st = image.read()
    
    # data in [rows, cols, bands]
    arr = np.moveaxis(st, 0, -1)
    
    if ax is None:
        ax = plt.gca() # get current axes
    
    nbands = np.array([bands]).flatten()
            
    # xlabel
    if label is None:
        label = {i:f'band {i}' for i in nbands}
    else:
        nlabel = np.array([label]).flatten()
        
        if not len(nbands) == len(nlabel):
            raise TypeError('Length of "bands" must be equal to "label".')
        
        label = {b:l for b,l in zip(nbands, nlabel)}
        
    for i in nbands:
        
        # Draw the plot
        arr_1d = arr[:, :, i].flatten()
        
        ax.hist(arr_1d, bins = bins, alpha = alpha, label = label[i], density = density, **kwargs)
    
    # title
    if title is None: title = "Histogram"
    
    # ylabel
    if ylabel is None: ylabel = "Frequency"
    
    # xlabel
    if xlabel is None: xlabel = "Band values"
        
    # Title and labels
    ax.set(title = title, xlabel = xlabel, ylabel = ylabel)

    plt.legend()
    plt.tight_layout()
    
    return ax


def plotRGB(image, bands = [5,4,3], stretch = 'std', title = None, xlabel = None, ylabel = None, 
            ax = None, **kwargs):
    
    '''
    Plotting a satellite image in RGB.
    
    This function allows to plot an satellite image in RGB channels.
        
    Parameters:
            
        image: Satellite image. It must be np.ndarray with bands, rows and cols or must be rasterio.io.DatasetReader with 3d.
            
        bands: A list contain the order of bands to be used in order to plot in RGB. For example,
                   for six bands (blue, green, red, nir, swir1 and swir2), number four (4) indicates 
                   the swir1 band, number three (3) indicates the nir band and the number two (2) indicates
                   the red band.
                   
        stretch: Contrast enhancement using the histogram. There are two options here: i) using
                     standard deviation ('std') and ii) using percentiles ('per'). For default is 'std', which means
                     standard deviation.
            
        title: Assigned title.
        
        xlabel: X axis title.
        
        ylabel: Y axis title.
        
        ax: current axes
        
        **kwargs: These will be passed to the matplotlib imshow(), please see full lists at:
                https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.imshow.html
    
    Return:
        ax: Graphic of an image in RGB.
        
    '''

    if isinstance(image, (np.ndarray)):
        # data in [rows, cols, bands]
        ts = np.moveaxis(image, 0, -1)

    elif isinstance(image, (rasterio.io.DatasetReader)):
        ts = image.read()
        # data in [rows, cols, bands]
        ts = np.moveaxis(ts, 0, -1)
    
    else:
        raise TypeError(f'{image} must be an array with 3d (bands, rows, cols) or must be raster read by rasterio.open().')
        
        
    if ts.shape[2] < 3:
        raise TypeError('"image" must be an image with at least three bands.')
    
    bands = bands
    
    arr_rgb = np.dstack([ts[:, :, (bands[0]-1)], ts[:, :, (bands[1]-1)], ts[:, :, (bands[2]-1)]])
    
    if stretch == 'std':
        
        mean = np.nanmean(arr_rgb)
        
        std = np.nanstd(arr_rgb)*1.5
        
        min_val = np.nanmax([mean - std, np.nanmin(arr_rgb)])
        
        max_val = np.nanmin([mean + std, np.nanmax(arr_rgb)])
        
        clipped_arr = np.clip(arr_rgb, min_val, max_val)
        
        arr_rgb_norm = (clipped_arr - min_val)/(max_val - min_val)

    elif stretch == 'per':
        
        p10 = np.nanpercentile(arr_rgb, 10) # percentile10
        
        p90 = np.nanpercentile(arr_rgb, 90) # percentile90
        
        clipped_arr = np.clip(arr_rgb, p10, p90)
        
        arr_rgb_norm = (clipped_arr - p10)/(p90 - p10)
        
    else:
        raise TypeError('Stretch type is not implemented. Please select either "std" or "per".')

    if ax is None:
        ax = plt.gca()

    ax.imshow(arr_rgb_norm, **kwargs)
    ax.grid(False)
    
    # title
    if title is not None:
        ax.set_title(title)
    
    # ylabel
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    
    # xlabel
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    
    #ax.set_axis_off()

    return ax


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# +
""" Tests for the scikit-eo package. """

import numpy as np
import pytest
import rasterio
import pandas as pd
from scikeo.process import confintervalML
from scikeo.sma import sma
from scikeo.pca import PCA
from scikeo.tassCap import tassCap
from scikeo.fusionrs import fusionrs
from scikeo.process import crop
from scikeo.rkmeans import rkmeans
from scikeo.mla import MLA
from scikeo.process import extract
from dbfread import DBF
from scikeo.calmla import calmla
from scikeo.deeplearning import DL


def test_confintervalML():
    
    """Confusion Matrix by Estimated Proportions of area an uncertainty is tested. 
    
    To carry out it, this function was tested using ground-truth values obtained by 
    Olofsson et al. (2014).
    """
    conf_error = pd.read_csv("tests/data/confusion_matrix.csv", index_col= 0, sep = ';')
    # only confusion matrix values
    values = conf_error.iloc[0:4,0:4].to_numpy()
    # number of pixels for each class
    img = np.array([200000, 150000, 3200000, 6450000])
    res = confintervalML(matrix = values, image_pred = img, pixel_size = 30, nodata = -9999)
    assert round(res.get('Overall_accuracy'),3) == 0.947
    
def test_sma():
    """Spectral mixture analysis is tested."""
    # image to be processed
    img = rasterio.open('tests/data/LC08_232066_20190727.tif')
    # endmembers
    endm =[[8980,8508,8704,13636,16579,11420], # soil
            [8207,7545,6548,16463,9725,6673], # forest
            [9276,9570,10089,6743,5220,5143], # water
           ]
    endm = np.array(endm)
    # applying the sma function
    frac = sma(image = img, endmembers = endm)
    assert frac.shape[2] == 3
    
def test_pca():
    """Principal Component Analysis is tested."""
    # image to be processed
    img = rasterio.open('tests/data/LC08_232066_20190727.tif')
    # Applying the PCA function:
    arr_pca = PCA(image = img, stand_varb = True)
    img_pca = arr_pca.get('PCA_image')
    assert img_pca.shape[2] == 6

def test_tassCap():
    """Tasseled-Cap is tested."""
    # image to be processed
    img = rasterio.open('tests/data/LC08_232066_20190727.tif')
    # Applying the tassCap function:
    arr_tct = tassCap(image = img, sat = 'Landsat8OLI')
    assert arr_tct.shape[2] == 3

def test_fusionrs():
    """Fusion optical and radar satelite images is tested."""
    # image to be processed
    # optical
    optical = rasterio.open('tests/data/LC08_003069_20180906_clip.tif')
    # radar
    radar = rasterio.open('tests/data/S1_2018_VV_VH_clip.tif')
    # Applying the fusionrs function:
    fusion = fusionrs(optical = optical, radar = radar)
    # Cumulative variance (%)
    cum_var = fusion.get('Cumulative_variance')*100
    assert round(cum_var[8],0) == 100

def test_crop():
    """cliping a satellite images is tested."""
    # raster to be clipped
    path_raster = "tests/data/LC08_232066_20190727.tif"
    # area of Interes -> shapefile
    path_shp = "tests/data/clip.shp"
    # Path where the image will be saved
    output_path_raster = "tests/data"
    # The raster name
    output_name = 'LC08_232066_20190727_clip'
    # Applying the crop() function:
    crop(image = path_raster, vector = path_shp, 
         filename = output_name, 
         filepath = output_path_raster)
    clip_image = rasterio.open(output_path_raster + '/' + output_name+ '.tif')
    assert type(clip_image) == rasterio.io.DatasetReader

def test_kmeans():
    """k-means classification is tested."""
    # image to be processed
    img = rasterio.open('tests/data/LC08_232066_20190727.tif')
    # Applying rkmeans() algorithm with four classes:
    arr_rkmeans = rkmeans(image = img, k = 4, max_iter = 300)
    assert type(arr_rkmeans) == np.ndarray

def test_machineLearning():
    """Machine learning algorithms are tested."""
    # image to be processed
    path_raster = "tests/data/LC08_232066_20190727.tif"
    img = rasterio.open(path_raster)
    # endmembers
    path_endm = "tests/data/endmembers.dbf"
    endm = DBF(path_endm)
    endm = pd.DataFrame(endm)
    # Instance of mla():
    inst = MLA(image = img, endmembers = endm)
    # Applying Random Forest with 60% of data to train
    rf_class = inst.RF(training_split = 0.6)
    assert rf_class.get('Overall_Accuracy') >= 0.7
    
def test_CalibratingMachineLearning():
    """Calibrating machine learning algorithms are tested."""
    # endmembers
    path_endm = "tests/data/endmembers.dbf"
    endm = DBF(path_endm)
    endm = pd.DataFrame(endm)
    # Instance of Calmla():
    inst = calmla(endmembers = endm)
    # Instance of splitData():
    data = inst.splitData()
    # Calibrating with Monte Carlo Cross-Validation Calibration (MCCV)
    error_mccv = inst.MCCV(split_data = data, models = ('svm', 'dt', 'rf', 'nb'), n_iter = 10)
    # error of Random Forest
    error_rf = np.mean(np.array(error_mccv.get('rf')))
    # error of Decision Tree
    error_dt = np.mean(np.array(error_mccv.get('dt')))
    assert error_dt > error_rf
    
def test_DeepLearning():
    """Deep Learning is tested."""
    # image to be processed
    path_raster = "tests/data/LC08_232066_20190727.tif"
    img = rasterio.open(path_raster)
    # endmembers
    path_endm = "tests/data/endmembers.dbf"
    endm = DBF(path_endm)
    endm = pd.DataFrame(endm)
    # Instance of DL():
    inst = DL(image = img, endmembers = endm)
    # Applying the FullyConnected() function of Deep Learning:
    fc = inst.FullyConnected(hidden_layers = 4, 
                         hidden_units = [64,16,8,8], 
                         output_units = 4,
                         input_shape = (6,), 
                         epochs = 100, 
                         batch_size = 32, 
                         training_split = 0.8)
    
    assert fc.get('Overall_Accuracy') >= 0.7



# -*- coding: utf-8 -*-
# +
import copy
import rasterio
import numpy as np
import pandas as pd
from dbfread import DBF
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier as DT
from sklearn.ensemble import RandomForestClassifier as RF
from sklearn.naive_bayes import GaussianNB as GNB
from sklearn.neural_network import MLPClassifier as MLP
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics import confusion_matrix
import warnings

class MLA(object):
    
    'Supervised classification in Remote Sensing'
    
    def __init__(self, image, endmembers, nodata = -99999):
        
        '''
        Parameter:
        
            image: Optical images. It must be rasterio.io.DatasetReader with 3d.
            
            endmembers: Endmembers must be a matrix (numpy.ndarray) and with more than one endmember. 
                    Rows represent the endmembers and columns represent the spectral bands.
                    The number of bands must be equal to the number of endmembers.
                    E.g. an image with 6 bands, endmembers dimension should be $n*6$, where $n$ 
                    is rows with the number of endmembers and 6 is the number of bands 
                    (should be equal).
                    In addition, Endmembers must have a field (type int or float) with the names 
                    of classes to be predicted.
            
            nodata: The NoData value to replace with -99999.
            
        '''
        
        self.image = image
        self.endm = endmembers
        self.nodata = nodata
        
        if not isinstance(self.image, (rasterio.io.DatasetReader)):
            raise TypeError('"image" must be raster read by rasterio.open().')
        
        bands = self.image.count
        
        rows = self.image.height
        
        cols = self.image.width
        
        st = image.read()
        
        # data in [rows, cols, bands]
        st_reorder = np.moveaxis(st, 0, -1) 
        # data in [rows*cols, bands]
        arr = st_reorder.reshape((rows*cols, bands))
        
        # nodata
        #if np.isnan(np.sum(arr)):
            #arr[np.isnan(arr)] = self.nodata
        
        # dealing with nan
        key_nan = np.isnan(np.sum(arr[:,0]))
        
        if key_nan:
            # saving un array for predicted classes
            class_final = arr[:, 0].copy()
            # positions with nan
            posIndx = np.argwhere(~np.isnan(class_final)).flatten()
            # replace np.nan -> 0
            arr[np.isnan(arr)] = 0
        
        # if it is read by pandas.read_csv()
        if isinstance(self.endm, (pd.core.frame.DataFrame)):
            
            for i in np.arange(self.endm.shape[1]):
                
                if all(self.endm.iloc[:,int(i)] < 100) & all(self.endm.iloc[:,int(i)] >= 0): indx = i; break
        
        # if the file is .dbf    
        elif isinstance(self.endm, (DBF)): # isinstance() function With Inheritance
            
            self.endm = pd.DataFrame(iter(self.endm))
            
            for i in np.arange(self.endm.shape[1]):
                
                if all(self.endm.iloc[:,int(i)] < 100) & all(self.endm.iloc[:,int(i)] >= 0): indx = i; break
        
        else:
            raise TypeError('"endm" must be .csv (pandas.core.frame.DataFrame).')

        if not self.endm.shape[1] == (bands + 1):
            raise ValueError('The number of columns of signatures (included the class column) must'
                             'be equal to the number of bands + 1.')
        
        self.indx = indx
        
        self.arr = arr
        
        self.rows = rows
        
        self.cols = cols
        
        if key_nan:
            self.key_nan = key_nan
            self.posIndx = posIndx
            self.class_final = class_final
        else:
            self.key_nan = key_nan
            
        
    def SVM(self, training_split = 0.8, random_state = None, kernel = 'linear', **kwargs):
        
        '''The Support Vector Machine (SVM) classifier is a supervised non-parametric statistical learning technique that 
        does not assume a preliminary distribution of input data. Its discrimination criterion is a 
        hyperplane that separates the classes in the multidimensional space in which the samples 
        that have established the same classes are located, generally some training areas.
        
        SVM support raster data read by rasterio (rasterio.io.DatasetReader) as input.
     
        
        Parameters:
    
            training_split: For splitting samples into two subsets, i.e. training data and for testing
                            data.
    
            kernel : {'linear', 'poly', 'rbf', 'sigmoid', 'precomputed'}, default='rbf' Specifies 
                     the kernel type to be used in the algorithm. It must be one of 'linear', 'poly', 
                     'rbf', 'sigmoid', 'precomputed' or a callable. If None is given, 'rbf' will 
                     be used. See https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html#sklearn.svm.SVC
                     for more details.
                     
            **kwargs: These will be passed to SVM, please see full lists at:
                  https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html#sklearn.svm.SVC
    
        Return:
        
            A dictionary containing labels of classification as numpy object, overall accuracy, kappa index, confusion matrix.
        '''
        
        # removing the class column
        X = self.endm.drop(self.endm.columns[[self.indx]], axis = 1)
        
        # only predictor variables
        y = self.endm.iloc[:, self.indx]
        
        # split in training and testing
        Xtrain, Xtest, ytrain, ytest = train_test_split(
            X.values, 
            y.values, 
            train_size = training_split, 
            test_size = 1 - training_split, 
            random_state = random_state)
        
        # applying a support vector machine
        inst_svm = SVC(kernel = kernel, **kwargs)
        
        # model trained
        mt_svm = inst_svm.fit(Xtrain, ytrain)
        
        labels_svm = mt_svm.predict(self.arr)
        
        # dealing with nan
        if self.key_nan:
            # image border like nan
            labels_svm = labels_svm[self.posIndx]
            self.class_final[self.posIndx] = labels_svm
            classSVM = self.class_final.reshape((self.rows, self.cols))
        else:
            classSVM = labels_svm.reshape((self.rows, self.cols))
        
        # Confusion matrix
        predic_Xtest = mt_svm.predict(Xtest)
        
        MC = confusion_matrix(ytest, predic_Xtest)
    
        MC = np.transpose(MC)

        # Users Accuracy and Commission
        total_rows = np.sum(MC, axis = 1)
        ua = np.diag(MC)/np.sum(MC, axis = 1)*100
        co = 100 - ua

        # Producer Accuracy and Comission
        total_cols = np.sum(MC, axis = 0)
        pa = np.diag(MC)/np.sum(MC, axis = 0)*100
        om = 100 - pa

        total_cols = np.concatenate([total_cols, np.repeat(np.nan, 3)])
        pa = np.concatenate([pa, np.repeat(np.nan, 3)])
        om = np.concatenate([om, np.repeat(np.nan, 3)])

        n = MC.shape[0]
        cm = np.concatenate([MC, total_rows.reshape((n,1)), ua.reshape((n,1)), 
                             co.reshape((n,1))], axis = 1)
        cm = np.concatenate([cm, total_cols.reshape((1,n+3)), pa.reshape((1,n+3)), 
                             om.reshape((1,n+3))])

        min_class = np.nanmin(labels_svm)
        max_class = np.nanmax(labels_svm)
        
        namesCol = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesCol.append(stri)

        namesCol.extend(['Total', 'Users_Accuracy', 'Commission'])

        namesRow = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesRow.append(stri)

        namesRow.extend(['Total', 'Producer_Accuracy', 'Omission'])
        namesRow

        confusionMatrix = pd.DataFrame(cm, 
                                        columns = namesCol, 
                                        index = namesRow)

        oa = accuracy_score(ytest, predic_Xtest)
        kappa = cohen_kappa_score(ytest, predic_Xtest)
        
        output = {'Overall_Accuracy': oa,
                  'Kappa_Index': kappa,
                  'Confusion_Matrix': confusionMatrix,
                  'Classification_Map': classSVM,
                  'Image': self.image
                 } 
    
        return output
    
    def DT(self, training_split = 0.8, random_state = None, **kwargs):
        
        '''Decision Tree is also a supervised non-parametric statistical learning technique, where the input data is divided recursively 
        into branches depending on certain decision thresholds until the data are segmented into homogeneous subgroups. 
        This technique has substantial advantages for remote sensing classification problems due to its flexibility, intuitive simplicity, 
        and computational efficiency.
        
        DT support raster data read by rasterio (rasterio.io.DatasetReader) as input.
        
        
        Parameters:
    
            training_split: For splitting samples into two subsets, i.e. training data and for testing
                            data.
                     
            **kwargs: These will be passed to DT, please see full lists at:
                  https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html#sklearn.tree.DecisionTreeClassifier
    
        Return:
        
            A dictionary containing labels of classification as numpy object, overall accuracy, kappa index, confusion matrix.
        '''
        
        # removing the class column
        X = self.endm.drop(self.endm.columns[[self.indx]], axis = 1)
            
        # only predictor variables
        y = self.endm.iloc[:, self.indx]
        
        # split in training and testing
        Xtrain, Xtest, ytrain, ytest = train_test_split(
            X.values, 
            y.values, 
            train_size = training_split, 
            test_size = 1 - training_split, 
            random_state = random_state)
        
        # applying a support vector machine
        inst_dt = DT(**kwargs)
        
        # model trained
        mt_dt = inst_dt.fit(Xtrain, ytrain)
        
        labels_dt = mt_dt.predict(self.arr)
        
        # dealing with nan
        if self.key_nan:
            # image border like nan
            labels_dt = labels_dt[self.posIndx]
            self.class_final[self.posIndx] = labels_dt
            classDT = self.class_final.reshape((self.rows, self.cols))
        else:
            classDT = labels_dt.reshape((self.rows, self.cols))
        
        # Confusion matrix
        predic_Xtest = mt_dt.predict(Xtest)
        
        MC = confusion_matrix(ytest, predic_Xtest)
    
        MC = np.transpose(MC)

        # Users Accuracy and Commission
        total_rows = np.sum(MC, axis = 1)
        ua = np.diag(MC)/np.sum(MC, axis = 1)*100
        co = 100 - ua

        # Producer Accuracy and Comission
        total_cols = np.sum(MC, axis = 0)
        pa = np.diag(MC)/np.sum(MC, axis = 0)*100
        om = 100 - pa

        total_cols = np.concatenate([total_cols, np.repeat(np.nan, 3)])
        pa = np.concatenate([pa, np.repeat(np.nan, 3)])
        om = np.concatenate([om, np.repeat(np.nan, 3)])

        n = MC.shape[0]
        cm = np.concatenate([MC, total_rows.reshape((n,1)), ua.reshape((n,1)), 
                             co.reshape((n,1))], axis = 1)
        cm = np.concatenate([cm, total_cols.reshape((1,n+3)), pa.reshape((1,n+3)), 
                             om.reshape((1,n+3))])

        min_class = np.nanmin(labels_dt)
        max_class = np.nanmax(labels_dt)
        
        namesCol = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesCol.append(stri)

        namesCol.extend(['Total', 'Users_Accuracy', 'Commission'])

        namesRow = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesRow.append(stri)

        namesRow.extend(['Total', 'Producer_Accuracy', 'Omission'])
        namesRow

        confusionMatrix = pd.DataFrame(cm, 
                                        columns = namesCol, 
                                        index = namesRow)

        oa = accuracy_score(ytest, predic_Xtest)
        kappa = cohen_kappa_score(ytest, predic_Xtest)
        
        output = {'Overall_Accuracy': oa,
                  'Kappa_Index': kappa,
                  'Confusion_Matrix': confusionMatrix,
                  'Classification_Map': classDT,
                  'Image': self.image
                 } 
    
        return output
    
    def RF(self, training_split = 0.8, random_state = None, **kwargs):
        
        '''Random Forest is a derivative of Decision Tree which provides an improvement over DT to overcome the weaknesses of a single DT. 
        The prediction model of the RF classifier only requires two parameters to be identified: the number of classification trees desired, 
        known as “ntree,” and the number of prediction variables, known as “mtry,” used in each node to make the tree grow.
        
        RF support raster data read by rasterio (rasterio.io.DatasetReader) as input.
        
        
        Parameters:
    
            training_split: For splitting samples into two subsets, i.e. training data and for testing
                            data.
                     
            **kwargs: These will be passed to RF, please see full lists at:
                  https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    
        Return:
        
            A dictionary containing labels of classification as numpy object, overall accuracy, kappa index, confusion matrix.
        '''
        
        # removing the class column
        X = self.endm.drop(self.endm.columns[[self.indx]], axis = 1)
            
        # only predictor variables
        y = self.endm.iloc[:, self.indx]
        
        # split in training and testing
        Xtrain, Xtest, ytrain, ytest = train_test_split(
            X.values, 
            y.values, 
            train_size = training_split, 
            test_size = 1 - training_split, 
            random_state = random_state)
        
        # applying a support vector machine
        inst_rf = RF(**kwargs)
        
        # model trained
        mt_rf = inst_rf.fit(Xtrain, ytrain)
        
        labels_rf = mt_rf.predict(self.arr)
        
        # dealing with nan
        if self.key_nan:
            # image border like nan
            labels_rf = labels_rf[self.posIndx]
            self.class_final[self.posIndx] = labels_rf
            classRF = self.class_final.reshape((self.rows, self.cols))
        else:
            classRF = labels_rf.reshape((self.rows, self.cols))
        
        # Confusion matrix
        predic_Xtest = mt_rf.predict(Xtest)
        
        MC = confusion_matrix(ytest, predic_Xtest)
    
        MC = np.transpose(MC)

        # Users Accuracy and Commission
        total_rows = np.sum(MC, axis = 1)
        ua = np.diag(MC)/np.sum(MC, axis = 1)*100
        co = 100 - ua

        # Producer Accuracy and Comission
        total_cols = np.sum(MC, axis = 0)
        pa = np.diag(MC)/np.sum(MC, axis = 0)*100
        om = 100 - pa

        total_cols = np.concatenate([total_cols, np.repeat(np.nan, 3)])
        pa = np.concatenate([pa, np.repeat(np.nan, 3)])
        om = np.concatenate([om, np.repeat(np.nan, 3)])

        n = MC.shape[0]
        cm = np.concatenate([MC, total_rows.reshape((n,1)), ua.reshape((n,1)), 
                             co.reshape((n,1))], axis = 1)
        cm = np.concatenate([cm, total_cols.reshape((1,n+3)), pa.reshape((1,n+3)), 
                             om.reshape((1,n+3))])

        min_class = np.nanmin(labels_rf)
        max_class = np.nanmax(labels_rf)
        
        namesCol = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesCol.append(stri)

        namesCol.extend(['Total', 'Users_Accuracy', 'Commission'])

        namesRow = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesRow.append(stri)

        namesRow.extend(['Total', 'Producer_Accuracy', 'Omission'])
        namesRow

        confusionMatrix = pd.DataFrame(cm, 
                                        columns = namesCol, 
                                        index = namesRow)

        oa = accuracy_score(ytest, predic_Xtest)
        kappa = cohen_kappa_score(ytest, predic_Xtest)
        
        output = {'Overall_Accuracy': oa,
                  'Kappa_Index': kappa,
                  'Confusion_Matrix': confusionMatrix,
                  'Classification_Map': classRF,
                  'Image': self.image
                 } 
    
        return output
    
    def NB(self, training_split = 0.8, random_state = None, **kwargs):
        
        '''Naive Bayes classifier is an effective and simple method for image classification based on probability theory. The NB 
        classifier assumes an underlying probabilistic model and captures the uncertainty about the model in a principled way, 
        that is, by calculating the occurrence probabilities of different attribute values for different classes in a training 
        set.
        
        NB support raster data read by rasterio (rasterio.io.DatasetReader) as input.
        
        
        Parameters:
    
            training_split: For splitting samples into two subsets, i.e. training data and for testing
                            data.
                     
            **kwargs: These will be passed to SVM, please see full lists at:
                  https://scikit-learn.org/stable/modules/generated/sklearn.naive_bayes.GaussianNB.html
    
        Return:
        
            A dictionary containing labels of classification as numpy object, overall accuracy, kappa index, confusion matrix.
        '''
        
        # removing the class column
        X = self.endm.drop(self.endm.columns[[self.indx]], axis = 1)
            
        # only predictor variables
        y = self.endm.iloc[:, self.indx]
        
        # split in training and testing
        Xtrain, Xtest, ytrain, ytest = train_test_split(
            X.values, 
            y.values, 
            train_size = training_split, 
            test_size = 1 - training_split, 
            random_state = random_state)
        
        # applying a support vector machine
        inst_nb = GNB(**kwargs)
        
        # model trained
        mt_nb = inst_nb.fit(Xtrain, ytrain)
        
        labels_nb = mt_nb.predict(self.arr)
        
        # dealing with nan
        if self.key_nan:
            # image border like nan
            labels_nb = labels_nb[self.posIndx]
            self.class_final[self.posIndx] = labels_nb
            classNB = self.class_final.reshape((self.rows, self.cols))
        else:
            classNB = labels_nb.reshape((self.rows, self.cols))
        
        # Confusion matrix
        predic_Xtest = mt_nb.predict(Xtest)
        
        MC = confusion_matrix(ytest, predic_Xtest)
    
        MC = np.transpose(MC)

        # Users Accuracy and Commission
        total_rows = np.sum(MC, axis = 1)
        ua = np.diag(MC)/np.sum(MC, axis = 1)*100
        co = 100 - ua

        # Producer Accuracy and Comission
        total_cols = np.sum(MC, axis = 0)
        pa = np.diag(MC)/np.sum(MC, axis = 0)*100
        om = 100 - pa

        total_cols = np.concatenate([total_cols, np.repeat(np.nan, 3)])
        pa = np.concatenate([pa, np.repeat(np.nan, 3)])
        om = np.concatenate([om, np.repeat(np.nan, 3)])

        n = MC.shape[0]
        cm = np.concatenate([MC, total_rows.reshape((n,1)), ua.reshape((n,1)), 
                             co.reshape((n,1))], axis = 1)
        cm = np.concatenate([cm, total_cols.reshape((1,n+3)), pa.reshape((1,n+3)), 
                             om.reshape((1,n+3))])

        min_class = np.nanmin(labels_nb)
        max_class = np.nanmax(labels_nb)
        
        namesCol = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesCol.append(stri)

        namesCol.extend(['Total', 'Users_Accuracy', 'Commission'])

        namesRow = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesRow.append(stri)

        namesRow.extend(['Total', 'Producer_Accuracy', 'Omission'])
        namesRow

        confusionMatrix = pd.DataFrame(cm, 
                                        columns = namesCol, 
                                        index = namesRow)

        oa = accuracy_score(ytest, predic_Xtest)
        kappa = cohen_kappa_score(ytest, predic_Xtest)
        
        output = {'Overall_Accuracy': oa,
                  'Kappa_Index': kappa,
                  'Confusion_Matrix': confusionMatrix,
                  'Classification_Map': classNB,
                  'Image': self.image
                 } 
    
        return output
    
    def NN(self, training_split = 0.8, max_iter = 300, random_state = None, **kwargs):
        
        '''This classification consists of a neural network that is organized into several layers, that is, an input layer of predictor 
        variables, one or more layers of hidden nodes, in which each node represents an activation function acting on a weighted input 
        of the previous layers’ outputs, and an output layer.
        
        NN support raster data read by rasterio (rasterio.io.DatasetReader) as input.
        
        
        Parameters:
    
            training_split: For splitting samples into two subsets, i.e. training data and for testing
                            data.
                     
            **kwargs: These will be passed to SVM, please see full lists at:
                  https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPClassifier.html
    
        Return:
        
            A dictionary containing labels of classification as numpy object, overall accuracy, kappa index, confusion matrix.
        '''
        
        # removing the class column
        X = self.endm.drop(self.endm.columns[[self.indx]], axis = 1)
            
        # only predictor variables
        y = self.endm.iloc[:, self.indx]
        
        # split in training and testing
        Xtrain, Xtest, ytrain, ytest = train_test_split(
            X.values, 
            y.values, 
            train_size = training_split, 
            test_size = 1 - training_split, 
            random_state = random_state)
        
        # applying neural network
        inst_nn = MLP(max_iter = max_iter, **kwargs)
        
        # model trained
        mt_nn = inst_nn.fit(Xtrain, ytrain)
        
        labels_nn = mt_nn.predict(self.arr)
        
        # dealing with nan
        if self.key_nan:
            # image border like nan
            labels_nn = labels_nn[self.posIndx]
            self.class_final[self.posIndx] = labels_nn
            classNN = self.class_final.reshape((self.rows, self.cols))
        else:
            classNN = labels_nn.reshape((self.rows, self.cols))
        
        # Confusion matrix
        predic_Xtest = mt_nn.predict(Xtest)
        
        MC = confusion_matrix(ytest, predic_Xtest)
    
        MC = np.transpose(MC)

        # Users Accuracy and Commission
        total_rows = np.sum(MC, axis = 1)
        ua = np.diag(MC)/np.sum(MC, axis = 1)*100
        co = 100 - ua

        # Producer Accuracy and Comission
        total_cols = np.sum(MC, axis = 0)
        pa = np.diag(MC)/np.sum(MC, axis = 0)*100
        om = 100 - pa

        total_cols = np.concatenate([total_cols, np.repeat(np.nan, 3)])
        pa = np.concatenate([pa, np.repeat(np.nan, 3)])
        om = np.concatenate([om, np.repeat(np.nan, 3)])

        n = MC.shape[0]
        cm = np.concatenate([MC, total_rows.reshape((n,1)), ua.reshape((n,1)), 
                             co.reshape((n,1))], axis = 1)
        cm = np.concatenate([cm, total_cols.reshape((1,n+3)), pa.reshape((1,n+3)), 
                             om.reshape((1,n+3))])

        min_class = np.nanmin(labels_nn)
        max_class = np.nanmax(labels_nn)
        
        namesCol = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesCol.append(stri)

        namesCol.extend(['Total', 'Users_Accuracy', 'Commission'])

        namesRow = []
        for i in np.arange(min_class, max_class + 1):
            stri = str(i)
            namesRow.append(stri)

        namesRow.extend(['Total', 'Producer_Accuracy', 'Omission'])
        namesRow

        confusionMatrix = pd.DataFrame(cm, 
                                        columns = namesCol, 
                                        index = namesRow)

        oa = accuracy_score(ytest, predic_Xtest)
        kappa = cohen_kappa_score(ytest, predic_Xtest)
        
        output = {'Overall_Accuracy': oa,
                  'Kappa_Index': kappa,
                  'Confusion_Matrix': confusionMatrix,
                  'Classification_Map': classNN,
                  'Image': self.image
                 } 
    
        return output
